'IDE-Voreinstellungen:
'Option Strict On
'Option Explicit On

'Projekt-Voreinstellungen:
'Imports System.Windows.Forms
'Imports Microsoft.VisualBasic.ControlChars

Imports System.Drawing
Imports System.Drawing.Drawing2D
Imports Microsoft.VisualBasic
Imports System.ComponentModel

<DesignerCategory("Code")> _
Public Class ChartControl : Inherits Control

   Private _PlotPoints(-1) As PointF
   Private _Route(-1) As PointF
   Private _StartPath As New Drawing2D.GraphicsPath
   Private _StartPen As System.Drawing.Pen
   Private _CurvePath As New Drawing2D.GraphicsPath
   Private _CurvePen As System.Drawing.Pen
   Private _PlotPath As New Drawing2D.GraphicsPath
   Private _PlotPen As System.Drawing.Pen
   Private _TextPath As New Drawing2D.GraphicsPath
   Private _TextBrush As New SolidBrush(Color.Blue)
   Private _GridPath As New Drawing2D.GraphicsPath
   Private _GridPen As System.Drawing.Pen

   Private _DisplayMatrix As New Drawing2D.Matrix

   Private _SFX As New StringFormat      'StringFormat der X-Achsen-Beschriftung
   Private _SFY As New StringFormat

   Private _GridRange As RectangleF       'Innenbereich des Charts

   Private _NoValueRange As New RectangleF(Single.MinValue, Single.MinValue, 0, 0)
   Private _ValueRange As RectangleF = _NoValueRange 'fasst minimale und maximale X / Y -Werte
   '                       der _Route in einer Struktur - entspricht also dem Chart- Innenbereich

   Private ReadOnly _GridCell20mm As New SizeF
   Private _DataError As String = "Keine Werte"

   Protected Overrides Sub Dispose(ByVal disposing As Boolean)
      If Not MyBase.IsDisposed Then
         'Da lohnt sich das Disposen mal so richtig!
         For Each D As IDisposable In New IDisposable() { _
                 _CurvePath, _CurvePen, _PlotPath, _PlotPen, _TextPath, _
                 _TextBrush, _GridPath, _GridPen, _DisplayMatrix, _SFX, _
                 _SFY}
            D.Dispose()
         Next
      End If
      MyBase.Dispose(disposing)
   End Sub

   Public Sub New()
      MyBase.DoubleBuffered = True                                                'mindert flackern
      Me.MinimumSize = New Size(50, 30)
      _GridPen = New Pen(Color.Black)
      _GridPen.DashPattern = New Single() {2, 4}
      _CurvePen = New Pen(Color.Red, 1)
      _StartPen = New Pen(Color.Red, 2)
      _PlotPen = New Pen(Color.FromArgb(0, 100, 0), 1.6)                           'sehr dunkles Grn

      '_PlotPen.DashPattern = New Single() {4, 2} 'dashed, damit die Plot-Kreuze ein Loch inne Mitte
      ''                                                                                      haben

      '_PlotPen.DashPattern = New Single() {4, 2} 'dashed, damit die Plot-Kreuze ein Loch inne Mitte
      ''                                                                                      haben

      Using G As Graphics = MyBase.CreateGraphics
         'anvisierte Gittergre: 20mm
         'Davon wird aber erheblich abgewichen, da die Minimierung der Nachkommastellen der
         ' Achsenbeschriftung wichtiger ist
         Const InchTo20mm As Single = 20 / 25.4F
         _GridCell20mm = New SizeF( _
             G.DpiX * InchTo20mm, G.DpiY * InchTo20mm)
      End Using
      MyBase.ForeColor = Color.Blue
      'X-Achsen-Beschriftungs-StringFormat 
      _SFX.Alignment = StringAlignment.Far  'richtet Text rechtsbndig am angegebenen Zielpunkt aus
      '                                                       (Text erscheint links vom Zielpunkt)
      _SFX.LineAlignment = StringAlignment.Center
      'Y-Achsen-Beschriftungs-StringFormat 
      _SFY.Alignment = StringAlignment.Center
      _SFY.LineAlignment = StringAlignment.Near   'richtet Text obenbndig am angegebenen Zielpunkt
      '                                                     aus  (Text erscheint unterm Zielpunkt)
   End Sub

   Public Sub ChangePoints(ByVal Route As IEnumerable(Of PointF), Optional ByVal plotPoints As PointF() = Nothing)
      If Route Is Nothing Then
         _Route = New PointF() {}
      Else
         _Route = Route.ToArray
      End If
      _PlotPoints = If(plotPoints, _Route)
      ApplyChanges()
   End Sub

   ''' <summary>
   ''' Ein Diagramm mit wenigen Punkten wird anders dargestellt
   ''' </summary>
   Private ReadOnly Property HasFewPoints() As Boolean
      Get
         Return _Route.Length < 3
      End Get
   End Property


   Private _GrowOnly As Boolean = False
   Public Property GrowOnly() As Boolean
      Get
         Return _GrowOnly
      End Get
      Set(ByVal value As Boolean)
         If _GrowOnly = value Then Return
         _GrowOnly = value
         ReGrow()
      End Set
   End Property

   Public ReadOnly ReGrow As Action = Sub() _ValueRange = _NoValueRange

   'bertrgt _Route nach _CurvePath. Dabei mu validiert werden
   Private Function BuildCurvePath() As Boolean
      Select Case _Route.Length
         Case 0
            _DataError = "Keine Werte"
         Case 1
            _DataError = String.Concat( _
               "Ein einzelner Punkt ", Lf, _
                _Route(0), Lf, " ist nicht im Diagramm darstellbar")
         Case Else
            '######## Hier wird die Kurve gebildet!! #######
            If HasFewPoints Then
               'als Spline-Kurve
               _CurvePath.AddCurve(_Route)
            Else
               'durch Geraden verbunden
               _CurvePath.AddLines(_Route)
            End If
            '########## Danke, das wars schon ############

            Dim rct = _CurvePath.GetBounds
            rct.Inflate(rct.Width / 4, rct.Height / 4)
            If _GrowOnly AndAlso _ValueRange <> _NoValueRange Then
               rct = RectangleF.Union(rct, _ValueRange)
            End If
            _ValueRange = rct
            With _ValueRange
               If .Height <> 0 AndAlso .Width <> 0 Then
                  _DataError = ""
                  Return True
               ElseIf .Equals(RectangleF.Empty) Then
                  _DataError = String.Concat( _
                      "Das Diagramm ist nicht darstellbar, weil alle", _
                      " Punkte auf derselben Position", _
                      Lf, _Route(0), " liegen.")
               Else
                  _DataError = String.Concat( _
                      "Der Wertebereich von ", Lf, .Location, _
                      " bis ", New PointF(.Right, .Bottom), Lf, _
                      " ist nicht im Diagramm darstellbar, da er", _
                      " keine flchige Ausdehnung hat.")
               End If

            End With
      End Select
      Return False
   End Function

   Private Sub ApplyChanges()
      _CurvePath.Reset()
      _PlotPath.Reset()
      _GridPath.Reset()
      _TextPath.Reset()
      Me.Invalidate()                                                     'Neuzeichnen anfordern...
      If Not BuildCurvePath() Then Return '... egal, ob hier abgebrochen wird oder nicht
      Dim _Scale As New SizeF( _
         _GridRange.Width / _ValueRange.Width, _
         _GridRange.Height / _ValueRange.Height)
      With _DisplayMatrix
         .Reset()
         .Translate(-_ValueRange.X, -_ValueRange.Y, MatrixOrder.Append)   'Wertebereich auf Nullpkt
         '                                                                                schieben
         .Scale(_Scale.Width, _Scale.Height, MatrixOrder.Append)           'auf GridRange skalieren
         .Translate(_GridRange.X, _GridRange.Y, MatrixOrder.Append)    'GridRange-Offset hinzufgen
         .Scale(1, -1, MatrixOrder.Append)     'ber die X-Achse spiegeln (also aussm Control raus)
         .Translate(0, MyBase.Height, MatrixOrder.Append)   'wieder in den sichtbaren Bereich holen
      End With
      BuildBackground(_Scale)
      _CurvePath.Transform(_DisplayMatrix)
      BuildPlotPath()
   End Sub

   Private Sub BuildPlotPath()
      'an jedem Punkt ein Kreuzchen malen
      If _PlotPoints.Length < 3 Then Return
      Dim Pts As PointF() = _PlotPoints.ToArray
      _DisplayMatrix.TransformPoints(Pts)
      With _PlotPath
         For Each Pt As PointF In Pts

            Const Offs As Single = 5

            'Const Offs As Single = 7.071

            '2 Diagonalen
            .StartFigure()
            .AddLine(Pt.X - Offs, Pt.Y - Offs, Pt.X + Offs, Pt.Y + Offs)
            .StartFigure()
            .AddLine(Pt.X - Offs, Pt.Y + Offs, Pt.X + Offs, Pt.Y - Offs)
         Next
      End With
      Pts = {_Route(0)}
      _DisplayMatrix.TransformPoints(Pts)
      Dim rct = New RectangleF(Pts(0), SizeF.Empty)
      rct.Inflate(4, 4)
      _StartPath.Reset()
      _StartPath.AddEllipse(rct)
   End Sub

   ''' <returns>Abstands-Werte fr Orientierungslinien</returns>
   Private Function GetSnapValue(ByVal Number As Double) As Single
      'sucht in der Zahlenfolge {... 0.1, 0.2, 0.5, 1, 2, 5, 10, ...} nach dem Number am nchsten
      ' kommenden Wert. Das ermglicht Achsenbschriftungen mit mglichst wenig Nachkommastellen
      'Int(Math.Log10(Val)): Komma-Position ( negativ fr: 0 < Val < 1 )
      Dim DecimalRangeBottom As Double = 10 ^ Int(Math.Log10(Number))
      Dim Numb As Double
      Dim Prev As Double = DecimalRangeBottom
      'testet das Doppelte, 5-fache, 10-fache
      For Each Factor As Double In New Double() {2, 5, 10}
         Numb = Factor * DecimalRangeBottom
         If Numb >= Number Then Exit For
         Prev = Numb
      Next
      If Numb - Number < Number - Prev Then
         Return CSng(Numb)
      Else
         Return CSng(Prev)
      End If
   End Function

   ''' <returns>Anzahl der Segmente, in die man Fullsize unterteilen kann</returns>
   Private Function AlignCount( _
         ByVal FullSize As Single, _
         ByVal SegmentSize As Single, _
         Optional ByVal UpperSide As Boolean = True) As Integer
      AlignCount = CInt(Int(FullSize / SegmentSize))
      ' Die Summe der Segmente (horizontal o. waagerecht) ist eigentlich immer kleiner als
      'FullSize. Auer, Value ist ganzzahlig teilbar durch SegmentSize.
      Dim Rest As Single = FullSize - AlignCount * SegmentSize
      If UpperSide Then
         'Gib eine Anzahl Segmente zurck, deren Summe >= Value
         If Rest > SegmentSize * 0.005F Then AlignCount += 1
      Else
         'Gib eine Anzahl Segmente zurck, deren Summe <= Value
         If Rest > SegmentSize * 0.995F Then AlignCount += 1
      End If
   End Function

   Private Sub BuildBackground(ByVal Scale As SizeF)
      'Das wre die Schrittweite der Werte in X- und Y-Richtung beim Beschriftungs-Abstand von 20mm
      Dim ValueStep20mm As New SizeF( _
         _GridCell20mm.Width / Scale.Width, _
         _GridCell20mm.Height / Scale.Height)

      'Wir suchen aber eine "runde" Schrittweite ( mglichst wenig Nachkommastellen )
      Dim ValueStep As New PointF( _
                GetSnapValue(ValueStep20mm.Width), _
                GetSnapValue(ValueStep20mm.Height))

      Dim Start As New Point( _
         AlignCount(_ValueRange.X, ValueStep.X), _
         AlignCount(_ValueRange.Y, ValueStep.Y))

      Dim [End] As New Point( _
         AlignCount(_ValueRange.Right, ValueStep.X, False), _
         AlignCount(_ValueRange.Bottom, ValueStep.Y, False))

      SetGridLines(Start, [End], ValueStep)
      SetGridCaptions(Start, [End], ValueStep)
   End Sub

   Private Sub SetGridLines(ByVal Start As Point, ByVal [End] As Point, ByVal ValueStep As PointF)
      With _GridPath
         For I As Integer = 0 To [End].X - Start.X
            Dim X As Single = (Start.X + I) * ValueStep.X
            .StartFigure()
            .AddLine(X, _ValueRange.Top, X, _ValueRange.Bottom)                          'Vertikale
         Next
         For I As Integer = 0 To [End].Y - Start.Y
            Dim Y As Single = (Start.Y + I) * ValueStep.Y
            .StartFigure()
            .AddLine(_ValueRange.Left, Y, _ValueRange.Right, Y)                        'Horizontale
         Next
         .AddRectangle(_ValueRange)                                                    'Rahmen drum
         .Transform(_DisplayMatrix)                        'Auf Gre bringen und vertikal spiegeln
      End With
   End Sub

   '_TextPath konstruieren
   Private Sub SetGridCaptions(ByVal Start As Point, ByVal [End] As Point, ByVal ValueStep As PointF)
      'X-Achse beschriften
      Dim UBound As Integer = [End].X - Start.X
      Dim Locations(UBound) As PointF
      Dim Entries(UBound) As Single
      ' Entries und Locations ermitteln
      For I As Integer = 0 To UBound
         Entries(I) = (I + Start.X) * ValueStep.X
         Locations(I) = New PointF(Entries(I), _ValueRange.Top)
      Next
      _DisplayMatrix.TransformPoints(Locations)                           'Locations transformieren
      ' an den Locations die Entries hinschreiben
      WriteAxis(Locations, Entries, _SFY)

      'Y-Achse beschriften
      UBound = [End].Y - Start.Y
      ReDim Locations(UBound)
      ReDim Entries(UBound)
      For I As Integer = 0 To UBound
         Entries(I) = (I + Start.Y) * ValueStep.Y
         Locations(I) = New PointF(_ValueRange.X, Entries(I))
      Next

      'Y-Achs-Beschriftung etwas nach links versetzen
      _DisplayMatrix.Translate(-7, 0, MatrixOrder.Append)
      _DisplayMatrix.TransformPoints(Locations)
      _DisplayMatrix.Translate(7, 0, MatrixOrder.Append)
      WriteAxis(Locations, Entries, _SFX)
   End Sub

   Private Sub WriteAxis( _
         ByVal Locations() As PointF, ByVal Entries() As Single, ByVal SF As StringFormat)
      For I As Integer = 0 To Locations.Length - 1
         With MyBase.Font
            _TextPath.AddString(Entries(I).ToString, .FontFamily, .Style, .Size, Locations(I), SF)
         End With
      Next
   End Sub

   Protected Overrides Sub OnFontChanged(ByVal e As System.EventArgs)
      MyBase.OnFontChanged(e)
      Me.Invalidate()
   End Sub

   Protected Overrides Sub OnForeColorChanged(ByVal e As System.EventArgs)
      MyBase.OnForeColorChanged(e)
      _TextBrush.Color = MyBase.ForeColor
      Me.Invalidate()
   End Sub

   Protected Overrides Sub OnPaint(ByVal e As PaintEventArgs)
      MyBase.OnPaint(e)
      Dim G As Graphics = e.Graphics
      If Not String.IsNullOrEmpty(_DataError) Then
         _TextPath.Reset()
         With MyBase.Font
            _TextPath.AddString( _
               String.Concat(MyBase.Text, Lf, _DataError), _
               .FontFamily, .Style, .Size, Point.Empty, _
                StringFormat.GenericDefault)
         End With
      Else
         G.DrawPath(_GridPen, _GridPath)     'Gitterlinien
         G.DrawPath(_PlotPen, _PlotPath)
         G.DrawPath(_CurvePen, _CurvePath)   'Kurve
         G.DrawPath(_StartPen, _StartPath)   'Kurve
      End If
      ControlPaint.DrawBorder( _
          G, Me.ClientRectangle, Color.Black, ButtonBorderStyle.Solid)
      G.SmoothingMode = SmoothingMode.AntiAlias     'Textdarstellung erfordert AntiAliasing 
      G.FillPath(_TextBrush, _TextPath)
   End Sub

   Protected Overrides Sub OnSizeChanged(ByVal e As EventArgs)
      MyBase.OnSizeChanged(e)
      Static PreventSizeZero As New Size(1, 1)
      'Gut zu wissen: ber die im Designer einstellbare Property MinimumSize wird der Nullpkt des
      ' Diagramm-Innenbereichs gesetzt.
      With MyBase.MinimumSize
         _GridRange = New RectangleF( _
            New Point(.Width - 10, .Height - 10), _
            PreventSizeZero + MyBase.Size - MyBase.MinimumSize)
      End With
      ApplyChanges()
   End Sub

End Class 'ChartControl
