﻿Imports System.Collections
Imports System.Linq
Imports Microsoft.VisualBasic

Public Class Heap(Of T)
   'ein Heap ist ein sortierter, balancierter binärer Baum, im Array (hier: List(of T)) abgebildet.
   'Die beiden Children eines Parents liegen genau hinter dem doppelten Index der Parents.
   'Heap-Bedingung (ascending): beide Children eines Parents müssen größer sein.

   Public ReadOnly Selector As Func(Of T, IComparable)
   Public ReadOnly Ascending As Boolean = True
   Public ReadOnly Flush As IEnumerable(Of T) = New HeapEnumerable(Of T)(Me)

   Public Sub New(ByVal ascending As Boolean, ByVal Selector As Func(Of T, IComparable))
      Me.Selector = Selector
      Me.Ascending = ascending
      If Ascending Then
         _Compare = AddressOf _CompareAsc
      Else
         _Compare = AddressOf _CompareDesc
      End If
   End Sub

   Public ReadOnly InnerList As New List(Of T)()
   Public Sub Clear()
      InnerList.Clear()
   End Sub

   Protected ReadOnly _Compare As Func(Of IComparable, T, Integer)

   Private Function _CompareDesc(ByVal value As IComparable, ByVal item As T) As Integer
      Return Selector(item).CompareTo(value)
   End Function

   Private Function _CompareAsc(ByVal value As IComparable, ByVal item As T) As Integer
      Return value.CompareTo(Selector(item))
   End Function

   Public Overridable Sub Push(ByVal item As T)
      InnerList.Add(Nothing)      ' leeren Platz am Ende schaffen
      'Durch "Aufrücken" wird die korrekte EinfügePosition für Value bereitgestellt
      'läuft, ausgehend von letzter Position, von Parent zu Parent, bis einer kleiner ist als Value
      Dim value = Selector(item)
      Dim iChild = InnerList.Count - 1
      While iChild <> 0
         Dim iParent = (iChild - 1) >> 1
         If _Compare(value, InnerList(iParent)) >= 0 Then Exit While 'Parent ist kleiner
         InnerList(iChild) = InnerList(iParent)         'bisherigen Parent auf iChild schreiben
         iChild = iParent
      End While
      InnerList(iChild) = item
   End Sub

   Public Function Pop() As T
      Dim retVal As T = InnerList(0)      'der erste, Rückgabewert - sein Platz wird frei
      Dim count = InnerList.Count - 1
      Dim homeless = InnerList(count) 'der letzte wird "obdachlos"
      Dim comp = Selector(homeless)
      InnerList.RemoveAt(count)
      If count = 0 Then Return retVal
      'es wird der Platz gesucht, wo homeless einzusetzen ist
      'Suche mittels Aufrücken: immer das kleinere der beiden Kinder übernimmt den Parent-Platz
      'Bis ein Platz gefunden ist, an dem homeless noch kleiner ist
      'Da mit jedem Loop sich die Schrittweite verdoppelt, geht das sehr schnell
      Dim iParent = 0
      Dim iMax = 0
      'Index des größeren der beiden Children
      Do
         ' Heap - binärer Baum im Array: die beiden Children eines Knotens
         ' liegen genau hinter der Verdopplung des Parent-Index'
         iParent = iMax
         iMax = (iParent << 1) + 1
         If iMax >= count Then Exit Do
         Dim otherMax = iMax + 1
         If otherMax < count AndAlso _
            _Compare(Selector(InnerList(otherMax)), InnerList(iMax)) < 0 Then iMax = otherMax
         If _Compare(comp, InnerList(iMax)) <= 0 Then Exit Do
         InnerList(iParent) = InnerList(iMax)
      Loop While True
      InnerList(iParent) = homeless
      Return retVal
   End Function

   Public Function Peek() As T
      Return InnerList(0)
   End Function

   Public ReadOnly Property Count() As Integer
      Get
         Return InnerList.Count
      End Get
   End Property

   Public Function CompareTopTo(ByVal value As IComparable) As Integer
      Return _Compare(value, InnerList(0))
   End Function



   ''' <summary>
   ''' bei Ascending: Ist das item größer als der TopNode, wird es aufgenommen, der TopNode returnt.
   ''' Nützlich, wenn man eine begrenzte Anzahl größtmöglicher Werte sucht. Bei Descending entsprechend.
   ''' </summary>
   Public Function TryExchange(ByRef item As T) As Boolean
      Dim comp = Selector(item).CompareTo(Selector(InnerList(0)))
      If comp = 0 OrElse Ascending = (comp < 0) Then Return False
      Dim itm = Pop()
      Push(item)
      item = itm
      Return True
   End Function

End Class
Public Class HeapEnumerable(Of T) : Inherits Enumerable(Of T)
   Private _Heap As Heap(Of T)
   Public Sub New(ByVal Heap As Heap(Of T))
      _Heap = Heap
   End Sub
   Protected Overrides Function TryMovenext(ByRef Current As T) As Boolean
      If _Heap.Count = 0 Then Return False
      Current = _Heap.Pop
      Return True
   End Function
End Class