﻿Imports System.Runtime.CompilerServices

Namespace System.Collections.Generic
   Public Module CollectionX

      Public Delegate Function GetChilds(Of T)(ByVal item As T) As IEnumerable

#Region "ICollection(ohne T)"

      <Extension()> _
      Public Function ToArray(Of T)(ByVal subj As System.Collections.ICollection) As T()
         ReDim ToArray(subj.Count - 1)
         If subj.Count > 0 Then subj.CopyTo(ToArray, 0)
      End Function

      <Extension()> _
      Public Sub ReplaceBy(Of T)(ByVal subj As ICollection(Of T), ByVal other As IEnumerable(Of T))
         subj.Clear()
         other.ForEach(AddressOf subj.Add)
      End Sub

#End Region 'ICollection(ohne T)

#Region "Array"

      <Extension()> _
      Public Function Clone(Of T)(ByVal subj As T()) As T()
         ReDim Clone(subj.Length - 1)
         Array.Copy(subj, Clone, subj.Length - 1)
      End Function

      <Extension()> _
      Public Sub CopyX(Of T)(ByVal Arr1() As T, ByVal Indx1 As Integer, ByVal Arr2() As T, ByVal Indx2 As Integer, ByVal Length As Integer)
         Array.Copy(Arr1, Indx1, Arr2, Indx2, Length)
      End Sub

      <Extension()> _
      Public Sub CopyX(Of T)(ByVal Arr1() As T, ByVal Arr2() As T, ByVal Indx2 As Integer, ByVal Length As Integer)
         Array.Copy(Arr1, 0, Arr2, Indx2, Length)
      End Sub

      <Extension()> _
      Public Sub CopyX(Of T)(ByVal Arr1() As T, ByVal Arr2() As T, ByVal Indx2 As Integer)
         Array.Copy(Arr1, 0, Arr2, Indx2, Arr1.Length)
      End Sub

      <Extension()> _
      Public Function FindIndex(Of T)(ByVal Arr() As T, ByVal IsMatch As Predicate(Of T)) As Integer
         Return Array.FindIndex(Arr, IsMatch)
      End Function

      <Extension()> _
      Public Function Find(Of T)(ByVal Arr() As T, ByVal IsMatch As Predicate(Of T)) As T
         Return Array.Find(Arr, IsMatch)
      End Function

      <Extension()> _
      Public Function IndexOf(Of T)(ByVal Arr() As T, ByVal Element As T) As Integer
         Return Array.IndexOf(Arr, Element)
      End Function

      <Extension()> _
      Public Function Contains(Of T)(ByVal Arr() As T, ByVal Element As T) As Boolean
         Return DirectCast(Arr, IList(Of T)).Contains(Element)
      End Function

      <Extension()> _
      Public Function ConvertAll(Of TSrc, TDest)( _
              ByVal Arr() As TSrc, ByVal Converter As Converter(Of TSrc, TDest)) As TDest()
         Return Array.ConvertAll(Arr, Converter)
      End Function

      <Extension()> _
      Public Sub KeepOnly(Of T)(ByRef subj() As T, ByVal predicate As Predicate(Of T))
         Dim i = subj.CloseUp(predicate)
         Dim newArr(i - 1) As T
         Array.Copy(subj, newArr, i)
         subj = newArr
      End Sub

      <Extension()> _
      Public Sub Sort(Of T)(ByVal subj() As T)
         Array.Sort(subj)
      End Sub

#End Region 'Array

#Region "List(Of T)"

      <Extension()> _
      Public Sub SortX(Of T)(ByVal Subj As List(Of T), ByVal ParamArray getters() As Func(Of T, IComparable))
         Dim comparison As Comparison(Of T)
         Select Case getters.Length
            Case 0 : Throw Subj.Exception("cannot SortX with no getter defined")
            Case 1
               Dim gt = getters(0)
               comparison = Function(x As T, y As T) gt(x).CompareTo(gt(y))
            Case Else
               comparison = _
                  Function(x As T, y As T)
                     For Each gt In getters
                        Dim cmp = gt(x).CompareTo(gt(y))
                        If cmp <> 0 Then Return cmp
                     Next
                     Return 0
                  End Function
         End Select
         Subj.Sort(comparison)
      End Sub

      <Extension()> _
      Public Sub KeepOnly(Of T)(ByRef subj As List(Of T), ByVal predicate As Predicate(Of T))
         Dim i = subj.CloseUp(predicate)
         subj.RemoveRange(i, subj.Count - i)
      End Sub

      <Extension()> _
      Public Sub InsertOrAdd(Of T)( _
            ByVal subj As List(Of T), ByVal item As T, ByVal predicate As Predicate(Of T))
         subj.InsertOrAdd(subj.FindIndex(predicate), item)
      End Sub

      <Extension()> _
      Public Sub InsertOrAt0(Of T)( _
            ByVal subj As List(Of T), ByVal item As T, ByVal predicate As Predicate(Of T))
         subj.InsertOrAt0(subj.FindIndex(predicate), item)
      End Sub

#End Region 'List(Of T)

#Region "Dictionary(Of T, T2)"

      <Extension()> _
      Public Function GetOrCreate(Of Tkey, Tvalue As New)( _
            ByVal subj As Dictionary(Of Tkey, Tvalue), ByVal key As Tkey) As Tvalue
         Dim ret As Tvalue
         If Not subj.TryGetValue(key, ret) Then
            ret = New Tvalue
            subj.Add(key, ret)
         End If
         Return ret
      End Function

#End Region 'Dictionary(Of T, T2)

#Region "IList"

      <Extension()> _
      Public Function CircIndex(ByVal subj As IList, ByVal index As Integer) As Integer
         index = index Mod subj.Count
         Return If(index < 0, index + subj.Count, index)
      End Function

      <Extension()> _
      Public Function CircAt(Of T)(ByVal subj As IList(Of T), ByVal index As Integer) As T
         index = index Mod subj.Count
         Return If(index < 0, subj(index + subj.Count), subj(index))
      End Function

      <Extension()> _
      Public Sub SetValues(Of T)(ByVal subj As IList(Of T), ByVal ParamArray items() As T)
         For i = 0 To items.Length - 1
            subj(i) = items(i)
         Next
      End Sub

      <Extension()> _
      Public Sub InsertOrAdd(Of T)(ByVal subj As IList(Of T), ByVal index As Integer, ByVal item As T)
         subj.Insert(If(index < 0, subj.Count, index), item)
      End Sub

      <Extension()> _
      Public Sub InsertOrAt0(Of T)(ByVal subj As IList(Of T), ByVal index As Integer, ByVal item As T)
         subj.Insert(If(index < 0, 0, index), item)
      End Sub

      <Extension()> _
      Public Function Clone(Of T As {IList, New})(ByVal subj As T) As T
         Clone = New T
         For Each itm In subj
            Clone.Add(itm)
         Next
      End Function

      ''' <summary>
      ''' Bsp: vorletztes Element: list.Last(-1)
      ''' </summary>
      <Extension()> _
      Public Function Last(Of T)(ByVal subj As IList(Of T), ByVal index As Integer) As T
         Return subj(index + subj.Count - 1)
      End Function

      <Extension()> _
      Public Function Pop(Of T)(ByVal Subj As List(Of T), Optional ByVal I As Integer = -1) As T
         If I < 0 Then
            I = Subj.Count - 1
            If I < 0 Then Return Nothing
         ElseIf I >= Subj.Count Then
            Return Nothing
         End If
         Pop = Subj(I)
         Subj.RemoveAt(I)
      End Function

      <Extension()> _
      Public Sub Swap(ByVal Subj As IList, ByVal I1 As Integer, ByVal I2 As Integer)
         Dim Tmp = Subj(I1)
         Subj(I1) = Subj(I2)
         Subj(I2) = Tmp
      End Sub

      <Extension()> _
      Public Sub RemoveRangeX(ByVal Subj As IList, ByVal start As Integer, Optional ByVal count As Integer = Integer.MaxValue)
         count.Minimize(Subj.Count - start)
         For i = start + count To Subj.Count - 1
            Subj(i - count) = Subj(i)
         Next
         For i = Subj.Count - 1 To Subj.Count - count Step -1
            Subj.RemoveAt(i)
         Next
      End Sub

#End Region 'IList

#Region "Enumerable"

      Public Structure MinMax(Of T)
         Public ReadOnly Min As T
         Public ReadOnly Max As T
         Public Sub New(ByVal Min As T, ByVal Max As T)
            Me.Min = Min : Me.Max = Max
         End Sub
      End Structure

      <Extension()> _
      Public Function Extremum(Of T, T2 As IComparable)(ByVal Subj As IEnumerable(Of T), ByVal Converter As Converter(Of T, T2)) As MinMax(Of T)
         With Subj.GetEnumerator
            If Not .MoveNext Then Throw New ArgumentException( _
               "In leerer Auflistung kann kein Extremum gefunden werden")
            Dim MinMax = New T() {.Current, .Current}
            Dim Val = Converter(.Current)
            Dim Values = New T2() {Val, Val}
            While .MoveNext
               Val = Converter(.Current)
               If Val.CompareTo(Values(0)) < 0 Then
                  MinMax(0) = .Current
                  Values(0) = Val
               ElseIf Val.CompareTo(Values(1)) > 0 Then
                  MinMax(1) = .Current
                  Values(1) = Val
               End If
            End While
            Return New MinMax(Of T)(MinMax(0), MinMax(1))
            .Dispose()
         End With
      End Function

      <Extension()> _
      Public Function AllChilds(Of T)( _
            ByVal GetChilds As Func(Of T, IEnumerable), _
            ByVal Roots As IEnumerable) As IEnumerable(Of T)
         Dim Enumerate As Func(Of IEnumerable(Of T), IEnumerable(Of T)) = Nothing
         Enumerate = Function(enmbl) enmbl.Concat(enmbl.SelectMany(Function(nd) Enumerate(GetChilds(nd).Cast(Of T))))
         Return Enumerate(Roots.Cast(Of T))
      End Function

      <Extension()> _
      Public Function All(Of T)( _
            ByVal GetChilds As Func(Of T, IEnumerable), ByVal Root As T) As IEnumerable(Of T)
         'Root wird in die Enumeration hineingenommen, indem ein es enthaltendes IEnumerable(Of T) 
         'geschaffen wird, dessen Kinder ( = nur Root ) rekursiv enumeriert werden
         Return GetChilds.AllChilds(New T() {Root})
      End Function

      <Extension()> _
      Public Function AllAsList(Of T)(ByVal getChilds As GetChilds(Of T), ByVal Roots As IEnumerable) As List(Of T)
         Dim ret As New List(Of T)
         For Each child As T In Roots : ret.Add(child) : Next
         Dim i As Integer = 0
         While i < ret.Count
            For Each child As T In getChilds(ret(i)) : ret.Add(child) : Next
            i += 1
         End While
         Return ret
      End Function

      <Extension()> _
      Public Function AllAsList(Of T)(ByVal getChilds As GetChilds(Of T), ByVal Root As T) As List(Of T)
         Return getChilds.AllAsList(New T() {Root})
      End Function

      <Extension()> _
      Public Function DisposeAll(Of T As IDisposable)(ByVal Subj As IList(Of T)) As IList(Of T)
         For Each O As T In Subj
            If O IsNot Nothing Then O.Dispose()
         Next
         If Not Subj.IsReadOnly Then Subj.Clear()
         Return Subj
      End Function

      <Extension()> _
      Public Sub DisposeAll(Of T As IDisposable)(ByVal Subj As IEnumerable(Of T))
         For Each O As T In Subj
            If O IsNot Nothing Then O.Dispose()
         Next
      End Sub

      <Extension()> _
      Public Function IndexOf(Of T)(ByVal Subj As IEnumerable(Of T), ByVal predicate As Func(Of T, Boolean)) As Integer
         For Each O In Subj
            If predicate(O) Then Exit Function
            IndexOf += 1
         Next
         Return -1
      End Function

      <Extension()> _
      Public Function IndexOf(Of T)(ByVal Subj As IList, ByVal predicate As Func(Of T, Boolean)) As Integer
         For Each O As T In Subj
            If predicate(O) Then Exit Function
            IndexOf += 1
         Next
         Return -1
      End Function

      <Extension()> _
      Public Sub ForEach(Of T)(ByVal Subj As IEnumerable, ByVal Action As Action(Of T))
         For Each O As T In Subj
            Action(O)
         Next
      End Sub

      <Extension()> _
      Public Sub ForEach(Of T)(ByVal Subj As IEnumerable(Of T), ByVal Action As Action(Of T))
         For Each O In Subj
            Action(O)
         Next
      End Sub

      <Extension()> _
      Public Function Except(Of T)(ByVal Subj As IEnumerable(Of T), ByVal item As T) As IEnumerable(Of T)
         Return Subj.Except(New T() {item})
      End Function

#End Region 'Enumerable

#Region "Helpers"

      <Extension()> _
      Private Function CloseUp(Of T)(ByVal subj As IList(Of T), ByVal predicate As Predicate(Of T)) As Integer
         Dim i As Integer
         For i = 0 To subj.Count - 1
            If Not predicate(subj(i)) Then Exit For
         Next
         For ii As Integer = i + 1 To subj.Count - 1
            If predicate(subj(ii)) Then
               subj(i) = subj(ii)
               i += 1
            End If
         Next
         Return i
      End Function

#End Region 'Helpers

   End Module

End Namespace
