﻿Namespace System.Drawing



   '<DebuggerDisplay("{ToString()}")> _

   <DebuggerDisplay("{Near}{Far}")> _
   Public Structure Rect : Implements IEnumerable(Of Vector)

      Public Near, Far As Vector
      Public Sub New(ByVal Near As Vector, ByVal Far As Vector)
         Me.Near = Near : Me.Far = Far
      End Sub
      Public Sub New(ByVal x1 As Integer, ByVal y1 As Integer, ByVal x2 As Integer, ByVal y2 As Integer)
         MyClass.New(New Vector(x1, y1), New Vector(x2, y2))
      End Sub
      Public Shared Function From(ByVal eachMember As Integer) As Rect
         Return Rect.From(Vector.From(eachMember))
      End Function
      Public Shared Function From(ByVal eachMember As Vector) As Rect
         Return New Rect(eachMember, eachMember)
      End Function
      Public Shared Function From(ByVal rct As Rectangle) As Rect
         Return New Rect(rct.Location, rct.Location + rct.Size)
      End Function
      Public Function Add(ByVal toEach As Integer) As Rect
         Return Add(Vector.From(toEach))
      End Function
      ''' <summary>performs offset </summary>
      Public Function Add(ByVal toEach As Vector) As Rect
         Return New Rect(Near + toEach, Far + toEach)
      End Function
      ''' <summary>
      ''' performs sophisticated offsets, eg. inflating/deflating (if v1 = -v2)
      ''' </summary>
      Public Function Add(ByVal v1 As Vector, ByVal v2 As Vector) As Rect
         Return New Rect(Near + v1, Far + v2)
      End Function
      ''' <summary>
      ''' performs sophisticated offsets, supports compatiblity to Rectangle
      ''' </summary>
      Public Function Add(ByVal r As Rect) As Rect
         Return New Rect(Near + r.Near, Far + r.Far)
      End Function
      Public Function Subtract(ByVal toEach As Integer) As Rect
         Return Subtract(Vector.From(toEach))
      End Function
      ''' <summary>performs offset </summary>
      Public Function Subtract(ByVal toEach As Vector) As Rect
         Return New Rect(Near - toEach, Far - toEach)
      End Function
      ''' <summary>
      ''' performs offset as well as inflating (use negative Vectors)
      ''' </summary>
      Public Function Subtract(ByVal v1 As Vector, ByVal v2 As Vector) As Rect
         Return New Rect(Near - v1, Far - v2)
      End Function
      ''' <summary>
      ''' performs sophisticated offsets, supports compatiblity to Rectangle
      ''' </summary>
      Public Function Subtract(ByVal r As Rect) As Rect
         Return New Rect(Near - r.Near, Far - r.Far)
      End Function
      Public Overrides Function ToString() As String
         Return " ".Between(Near, Far)
      End Function
      ''' <remarks>the getter is for compatibility only. better access Rect.Near directly</remarks>
      Public Property Location() As Vector
         Get
            Return Near
         End Get
         Set(ByVal value As Vector)
            Far += value - Near : Near = value
         End Set
      End Property
      Public Property Size() As Vector
         Get
            Return Far - Near
         End Get
         Set(ByVal value As Vector)
            Far = Near + value
         End Set
      End Property
      Default Public Property Item(ByVal far As Integer) As Vector
         Get
            If far = 0 Then Return Near
            Return Me.Far
         End Get
         Set(ByVal value As Vector)
            If far = 0 Then Near = value Else Me.Far = value
         End Set
      End Property
      Default Public Property Item(ByVal far As Boolean) As Vector
         Get
            Return If(far, Me.Far, Near)
         End Get
         Set(ByVal value As Vector)
            If far Then Me.Far = value Else Near = value
         End Set
      End Property
      Default Public Property Item(ByVal far As Integer, ByVal vertical As Integer) As Integer
         Get
            If far = 0 Then Return Near(vertical)
            Return Me.Far(vertical)
         End Get
         Set(ByVal value As Integer)
            If far = 0 Then Near(vertical) = value Else Me.Far(vertical) = value
         End Set
      End Property
      ''' <summary>
      ''' a none-intersection is indicated by returning a imaginary Rect
      ''' </summary>
      Public Function Intersect(ByVal other As Rect) As Rect
         For i = 0 To 1
            Intersect.Near(i) = Near(i).Max(other.Near(i))
            Intersect.Far(i) = Far(i).Min(other.Far(i))
         Next
      End Function
      ''' <summary>true, if my size has a negative component</summary>
      Public ReadOnly Property IsImaginary() As Boolean
         Get
            If Far.X - Near.X < 0 Then Return True
            Return Far.Y - Near.Y < 0
         End Get
      End Property
      Public Shared Operator -(ByVal r As Rect) As Rect
         Return New Rect(-r.Near, -r.Far)
      End Operator
      Public Shared Operator =(ByVal v1 As Rect, ByVal v2 As Rect) As Boolean
         Return v1.Near = v2.Near AndAlso v1.Far = v2.Far
      End Operator
      Public Shared Operator <>(ByVal v1 As Rect, ByVal v2 As Rect) As Boolean
         Return Not v1 = v2
      End Operator
      Public Overrides Function Equals(ByVal obj As Object) As Boolean
         If TypeOf obj Is Rect OrElse TypeOf obj Is Rectangle Then Return Me = CType(obj, Rect)
      End Function
      Public Overrides Function GetHashCode() As Integer
         Return Near.GetHashCode Xor Far.GetHashCode
      End Function
      Public Shared Widening Operator CType(ByVal other As Rectangle) As Rect
         Return New Rect(other.Location, other.Location + other.Size)
      End Operator
      Public Shared Widening Operator CType(ByVal other As Rect) As Rectangle
         Return New Rectangle(other.Near, other.Far - other.Near)
      End Operator
      Public Overloads Function EachEquals(ByVal amount As Integer) As Boolean
         Return Near.EachEquals(amount) AndAlso Far.EachEquals(amount)
      End Function
      ''' <summary>enables Linq-power to Rects elements</summary>
      Public Function GetEnumerator() As IEnumerator(Of Vector) Implements _
            IEnumerable(Of Vector).GetEnumerator
         Return DirectCast(New Vector() {Near, Far}, IEnumerable(Of Vector)) _
            .GetEnumerator
      End Function

      Private Function GetEnumerator1() As IEnumerator Implements _
            IEnumerable.GetEnumerator
         Return DirectCast(New Vector() {Near, Far}, IEnumerable(Of Vector)) _
            .GetEnumerator
      End Function
   End Structure

End Namespace
